local ConfigService = require("api/ConfigService")
local all_modems = require("vuci.modem"):get_all_modems()
local util = require("vuci.util")
local event_type = require("vuci.event_type")

local EventsReporting = ConfigService:new({
	anonymous = true
})

local ERR_CODES = {
	EMAIL_SEND_FAILED = 1,
	EMAIL_GROUP_NOT_FOUND = 2
}

function EventsReporting:send_test_email(email_data)
	local v = require "lualibparam"
	local data = email_data or self.arguments.data

	local all_recipients = ""
	if type(data.recipients) == "table" then
		all_recipients = table.concat(data.recipients, " ")
	else
		all_recipients = data.recipients
	end
	all_recipients = util.shellquote(all_recipients)

	if string.find(data.message, "%%et") then
		data.message = string.gsub(data.message, "%%et", data.event)
	end
	if string.find(data.message, "%%ex") then
		data.message = string.gsub(data.message, "%%ex", "Test email")
	end
	local result = v.expand_params(data.message, data.info_modem_id)
	data.message = util.trim(result)

	local group
	self:table_foreach("user_groups", "email", function(s)
		if s.name == data.group then
			group = s
		end
	end)

	group.secure_conn = group.secure_conn or "0"
	if group.secure_conn == "1" then
		if group.credentials == "1" then
			group.username = " -au" .. util.shellquote(group.username)
			group.password = " -ap" .. util.shellquote(group.password)
		else
			group.username = " "
			group.password = " "
		end
		util.exec("(echo -e "..util.shellquote("subject:"..data.subject.."\nfrom:"..group.senderemail.."\n"..data.message)
		.." | sendmail -v -H "..util.shellquote("exec openssl s_client -quiet -connect "..util.shellquote(group.smtp_ip..":"..group.smtp_port)
		.." -tls1_2 -starttls smtp").." -f "..util.shellquote(group.senderemail).." "..group.username.." "..group.password
		.." "..all_recipients.."; echo $? >/tmp/sendmail )")
	elseif group.secure_conn == "0" then
		if group.credentials == "1" then
			group.username = " -au"..util.shellquote(group.username)
			group.password = " -ap"..util.shellquote(group.password)
		else
			group.username = " "
			group.password = " "
		end
		util.exec("(echo -e "..util.shellquote("subject:"..data.subject.."\nfrom:"..group.senderemail.."\n"..data.message)
		.." | sendmail -S "..util.shellquote(group.smtp_ip..":"..group.smtp_port).." -f "..util.shellquote(group.senderemail)
		.." "..group.username.." "..group.password.." "..all_recipients.."; echo $? >/tmp/sendmail )")
	end

	local f = io.open("/tmp/sendmail", "r")
	if not f then
		self:add_critical_error(ERR_CODES.EMAIL_SEND_FAILED, "Failed to send an email")
	else
		local line = f:read()
		f:close()
		if line then
			if util.trim(line) == "0" then
				self:ResponseOK()
			end
		end
		self:add_critical_error(ERR_CODES.EMAIL_SEND_FAILED, "Failed to send an email")
	end
end

function EventsReporting:GET_TYPE_options()
	self:ResponseOK({
		events = event_type:get_all_events()
	})
end

local opt_enable

function EventsReporting:require_validation()
	local enable = self:get_abs_value(self.config, self.sid, "enable") or self.current_data_block["enable"]
	local _require = { "event", "eventMark", "action" }
	local main_require_options = {}
	local additional_require_options = {}

	if enable == "1" then
		local action = self:get_abs_value(self.config, self.sid, "action") or self.current_data_block["action"]
		if all_modems and #all_modems > 1 then
			if action and action == "sendEmail" then
				additional_require_options = {"info_modem_id"}
			elseif action and action == "sendSMS" then
				additional_require_options = {"send_modem_id", "info_modem_id"}
			end
		end

		if action and action == "sendEmail" then
			main_require_options = { "subject", "message", "emailgroup", "recipEmail"}
		elseif action and action == "sendSMS" then
			local recipient_require = nil
			local recipient_format_value = self:get_abs_value(self.config, self.sid, "recipient_format") or self.current_data_block["recipient_format"]

			if recipient_format_value and recipient_format_value == "single" then
				recipient_require = "telnum"
			elseif recipient_format_value and recipient_format_value == "group" then
				recipient_require = "group"
			end
			main_require_options = {"message", "recipient_format", recipient_require}
		end

		for _, v in ipairs(additional_require_options) do
			table.insert(_require, v)
		end
		for _, v in ipairs(main_require_options) do
			table.insert(_require, v)
		end

		opt_enable.require = { ["1"] = _require }
	end
end

EventsReporting.PUT_validate_section_hook = EventsReporting.require_validation
EventsReporting.POST_validate_section_hook = EventsReporting.require_validation

local Rules = EventsReporting:section("events_reporting", "rule")
function Rules:filter(options)
	return options.action ~= "sendRMS"
end

	opt_enable = Rules:option("enable")
		function opt_enable:validate(value)
			return self.dt:is_bool(value)
		end

	local opt_event = Rules:option("event")
		function opt_event:validate(value)
			local events = {}
			for event in pairs(event_type:get_all_events()) do
				table.insert(events, event)
			end
			return self.dt:check_array(value, events)
		end

	local opt_eventMark = Rules:option("eventMark")
		function opt_eventMark:validate(value)
			local current_event = self:get_abs_value(self.config, self.sid, "event")
			if not current_event then
				return false, "event not set"
			end
			local events = event_type:get_all_events()
			if not events[current_event] then
				return false, "event type not found"
			end
			return self.dt:check_array(value, events[current_event])
		end
		function opt_eventMark:get()
			local event = self:table_get(self.config, self.sid, "event")
			local event_mark = self:table_get(self.config, self.sid, "eventMark")
			if event and event_mark and event == "Reboot" and event_mark == "sms" then
				return "sms reboot"
			end
			return event_mark
		end
		function opt_eventMark:set(value)
			local event = self:get_abs_value(self.config, self.sid, "event")
			if event and value and event == "Reboot" and value == "sms reboot" then
				self:table_set(self.config, self.sid, "eventMark", "sms")
			else
				self:table_set(self.config, self.sid, "eventMark", value)
			end
		end

	local opt_action = Rules:option("action")
		function opt_action:validate(value)
			local check = {"sendEmail"}
			if all_modems and #all_modems > 0 then
				table.insert(check, "sendSMS")
			end
			return self.dt:check_array(value, check)
		end

	local opt_info_modem_id = Rules:option("info_modem_id")
		function opt_info_modem_id:validate(value)
			if all_modems and #all_modems > 1 then
				return self.dt:check_modem(value)
			end
			return false, "info_modem_id is available only if device has more than one modem"
		end

	local opt_send_modem_id = Rules:option("send_modem_id")
		function opt_send_modem_id:validate(value)
			if all_modems and #all_modems > 1 then
				return self.dt:check_modem(value)
			end
			return false, "send_modem_id is available only if device has more than one modem"
		end

	local opt_subject = Rules:option("subject")
		opt_subject.maxlength = 256
		function opt_subject:validate(value)
			return self.dt:fieldvalidation(value, "^[a-zA-Z0-9!@#$%%&*+/=?^_`{|}~%. %-]+$")
		end

	local opt_message = Rules:option("message")
		function opt_message:validate(value)
			return self.dt:string(value)
		end

	local opt_recipient_format = Rules:option("recipient_format")
		function opt_recipient_format:validate(value)
			return self.dt:check_array(value, {
				"single", "group"
			})
		end

	local opt_telnum = Rules:option("telnum")
		-- Disabled till WebUI fixes its creation of empty configurations.
		-- opt_telnum.cfg_require = true
		function opt_telnum:validate(value)
			return self.dt:phonedigit(value)
		end

	local opt_group = Rules:option("group")
		function opt_group:validate(value)
			local ok = false
			self:table_foreach("user_groups", "phone", function (s)
				if s.name == value then ok = true end
			end)
			return ok, "phone group not found"
		end

	local opt_emailgroup = Rules:option("emailgroup")
		function opt_emailgroup:validate(value)
			local ok = false
			self:table_foreach("user_groups", "email", function (s)
				if s.name == value then ok = true end
			end)
			return ok, "email group not found"
		end

	local opt_recipEmail = Rules:option("recipEmail", {list = true})
		-- Disabled till WebUI fixes its creation of empty configurations.
		-- opt_recipEmail.cfg_require = true
		function opt_recipEmail:validate(value)
			return self.dt:email(value)
		end

local EmailTest = EventsReporting:action("send_test_email", function (self, data)
	self:send_test_email(data)
end)

	local aopt_event = EmailTest:option("event")
		aopt_event.require = true
		aopt_event.validate = opt_event.validate

	local aopt_subject = EmailTest:option("subject")
		aopt_subject.require = true
		aopt_subject.validate = opt_subject.validate

	local aopt_message = EmailTest:option("message")
		aopt_message.require = true
		aopt_message.validate = opt_message.validate

	local aopt_recipients = EmailTest:option("recipients")
		aopt_recipients.require = true
		function aopt_recipients:validate(value)
			if type(value) == "table" then
				for _, v in ipairs(value) do
					return self.dt:email(v)
				end
			end
			return self.dt:email(value)
		end

	local aopt_group = EmailTest:option("group")
		aopt_group.require = true
		aopt_group.validate = opt_emailgroup.validate

	local aopt_info_modem_id = EmailTest:option("info_modem_id")
		function aopt_info_modem_id:validate(value)
			if all_modems and #all_modems > 1 then
				return self.dt:check_modem(value)
			end
			return false, "info_modem_id is available only if device has more than one modem"
		end

function EventsReporting:POST_validate_hook()
	local interfaces = 0
	self:table_foreach("events_reporting", "rule", function (_)
		interfaces = interfaces + 1
	end)
	if interfaces >= 90 then
		self:add_critical_error(STD_CODES.UCI_CREATE_ERROR, "Can't create more instances. Only 90 instances are allowed")
	end
end

return EventsReporting