import sys
sys.path.append("../../../../tests")
from utility_integration import WrapTest, Env
from utils.ssh import open_ssh_connection, send_cmd
from utils.general_api import get_hwinfo
from utils.mobile_utilities import get_io_pins, enable_scheduler_general, create_scheduler_instance, delete_scheduler_instance, disable_all_sms_rules, enable_all_sms_rules
import json
import response_codes as codes

RC = codes.ResponseCodes

class CallUtilities(WrapTest):
    url_general = "/services/mobile_utilities/call_utilities/general/config/general"
    url_action = "/services/mobile_utilities/call_utilities/rules/options"
    url_rules = "/services/mobile_utilities/call_utilities/rules/config"
    url_scheduler = "/services/io/scheduler/config/general/instances"
    pins = []
    scheduler_instances = []

    @classmethod
    def setUpClass(cls):
        cls.output_gpio_pins, cls.output_relay_pins = get_io_pins()

    def save_scheduler_instances(self):
        x = self.get(self.url_scheduler)
        self.scheduler_instances.extend(x.resp.json()["data"])
        for instance in self.scheduler_instances:
            if "enabled" in instance and instance["enabled"] == "1":
                x = self.put_data(f'{self.url_scheduler}/{instance["id"]}', {
                    "enabled": "0"
                })

    def revert_scheduler_instances(self):
        for instance in self.scheduler_instances:
            if "enabled" in instance and instance["enabled"] == "1":
                x = self.put_data(f'{self.url_scheduler}/{instance["id"]}', {
                    "enabled": "1"
                })

    def test_call_general(self):
        actions = ["reject", "ignore", "answer"]

        with self.subTest("Clear section"):
            x = self.put_data(self.url_general, {
                "action": "",
                "line_close_time": ""
            })
            x.assert_data({
                "id": "general",
                ".type": "call"
            }, 200)

        with self.subTest("Check all actions"):
            for action in actions:
                if action == "answer":
                    x = self.put_data(self.url_general, {
                        "action": action,
                        "line_close_time": "5"
                    })
                    x.assert_data({
                        "action": action,
                        "line_close_time": "5",
                        "id": "general",
                        ".type": "call"
                    }, 200)
                else:
                    x = self.put_data(self.url_general, {
                        "action": action,
                    })
                    x.assert_data({
                        "action": action,
                        "id": "general",
                        ".type": "call"
                    }, 200)

    def test_call_rules_action(self):
        actions = ["reboot", "send_status", "mobile"]
        hwinfo = get_hwinfo(self)
        
        if hwinfo["ios"]:
            actions.append("dout")

        if hwinfo["wifi"]:
            actions.append("wifi")

        if len(self.output_relay_pins) >= 1:
            actions.append("relay")

        x = self.get(self.url_action)
        x.assert_data({
            "actions": actions 
        })

    def test_call_rules_pin_option(self):
        if not get_hwinfo(self)["ios"]:
            self.skipTest("for this test need pin in device")
        
        with self.subTest("Check 'pin' option with all gpio pins"):
            if len(self.output_gpio_pins) > 0:
                x = self.post_data(self.url_rules, {
                    "action": "dout"
                })
                sid = x.resp.json()['data']['id']
                x.assert_data({
                    "value": "0",
                    ".type": "rule",
                    "action": "dout",
                    "id": sid
                }, 201)

                for pin in self.output_gpio_pins:
                    x = self.put_data(f'{self.url_rules}/{sid}', {
                        "pin": pin
                    })
                    x.assert_data({
                        ".type": "rule",
                        "action": "dout",
                        "value": "0",
                        "pin": pin,
                        "id": sid
                    }, 200)

                x = self.delete(f'{self.url_rules}/{sid}')
                x.assert_data({
                    "id": sid
                })

        with self.subTest("Check 'pin' option with all relay pins"):
            if len(self.output_relay_pins) > 0:
                x = self.post_data(self.url_rules, {
                    "action": "relay"
                })
                sid = x.resp.json()['data']['id']
                x.assert_data({
                    "value": "0",
                    ".type": "rule",
                    "action": "relay",
                    "id": sid
                }, 201)

                for pin in self.output_relay_pins:
                    x = self.put_data(f'{self.url_rules}/{sid}', {
                        "pin": pin
                    })
                    x.assert_data({
                        ".type": "rule",
                        "action": "relay",
                        "value": "0",
                        "pin": pin,
                        "id": sid
                    }, 200)

                x = self.delete(f'{self.url_rules}/{sid}')
                x.assert_data({
                    "id": sid
                })
        
    def test_call_rules_pin_using_scheduler(self):

        if not get_hwinfo(self)["ios"]:
            self.skipTest("for this test need pin in device")

        with self.subTest("Disable all scheduler instances and save them"):
            self.save_scheduler_instances()

        with self.subTest("Disable all SMS rules"):
            disable_all_sms_rules(self)

        enable_scheduler_general(self)

        if len(self.output_gpio_pins) > 0:

            with self.subTest("Create scheduler instance"):
                scheduler_sid = create_scheduler_instance(self, 1, self.output_gpio_pins[0])

            with self.subTest("Create call utilities instance"):
                x = self.post_data(self.url_rules, {
                    "enabled": "1",
                    "action": "dout",
                    "pin": self.output_gpio_pins[0]
                })
                x.assert_error("enabled", f"Unable to enable. Output scheduler instance with '{self.output_gpio_pins[0]}' pin is enabled", 1)

            with self.subTest("Delete scheduler instance"):
                delete_scheduler_instance(self, scheduler_sid)

        if len(self.output_relay_pins) > 0:
            with self.subTest("Create scheduler instance"):
                scheduler_sid = create_scheduler_instance(self, 1, self.output_relay_pins[0])

            with self.subTest("Create call utilities instance"):
                x = self.post_data(self.url_rules, {
                    "enabled": "1",
                    "action": "relay",
                    "pin": self.output_relay_pins[0]
                })
                x.assert_error("enabled", f'Unable to enable. Output scheduler instance with \'{self.output_relay_pins[0]}\' pin is enabled', 1)

            with self.subTest("Delete scheduler instance"):
                delete_scheduler_instance(self, scheduler_sid)

        with self.subTest("Enable all SMS rules"):
            enable_all_sms_rules(self)

        with self.subTest("Revert scheduler instances 'enabled' option values"):
            self.revert_scheduler_instances()