import io
import os
import re
import requests
from utils.ssh import get_ssh
import utility_integration as util
import sys
Env = util.Env
http = Env.http
sys.path.append("../../../../tests")

BASE_CFG_URL = "/services/packages/config"
INSTALL_PKG_URL = "/services/packages/actions/install_package"
UPDATE_PKG_URL = "/services/packages/actions/update_package"
DELETE_PKG_FILES_URL = "/services/packages/actions/delete_install_files"
BASE_STATUS_URL = "/services/packages/status"
TEST_PKG = "azure_iot_hub"
TEST_PKG_IPK_NAME = "azure_iothub"
TEST_PKG_TAR = TEST_PKG + ".tar.gz"
TEST_PKG_PATH = "./files/" + TEST_PKG_TAR
TEST_INVALID_PKG_PATH = "./files/" + "invalid_" +TEST_PKG_TAR
TEST_PKG_DIR = "./files/" + TEST_PKG
PKG_REBOOT = False
PKG_NETWORK_RESTART = False

class test_packages(util.WrapTest):
    @classmethod
    def setUpClass(self):
        self.internet_connection = False
        self.ssh = get_ssh()
        res = self.ssh.send_cmd("ping -c 1 1.1.1.1 &> /dev/null ; echo $?")
        self.internet_connection = res.strip() == "0"

        self.get_opkg_info()
        self.get_invalid_test_pkg()
        self.fake_versions()

        if self.internet_connection:
            # fetch new pkg list from server
            self.fetch_server_packages()

    @classmethod
    def tearDownClass(self):
        self.restore_versions()

        if self.internet_connection:
            # fetch current pkg list from server
            self.fetch_server_packages()

        self.ssh.logout()

    @classmethod
    def fetch_server_packages(self):
        self.ssh.send_cmd("opkg --force_feeds /etc/opkg/teltonikafeeds.conf -f /etc/tlt_opkg.conf update")
        self.ssh.prompt()
        self.ssh.prompt()

    def pkg_structure_test(self, pkg, url):
        required_keys = ["type", "Description", "Package", "tlt_name"]

        if "installed" not in url:
            required_keys.append("Filename")
            required_keys.append("Depends")
            required_keys.append("Size")
            required_keys.append("SHA256sum")
            required_keys.append("Installed-Size")
            required_keys.append("Architecture")
            required_keys.append("Version")
            required_keys.append("Router")
            required_keys.append("Section")
            required_keys.append("Firmware")

            self.assertEqual(self.release_fw_version, pkg["Firmware"])
            self.assertIn(self.platform, pkg["Router"])
            self.assertTrue(re.search("^[0-9]+$", pkg["Installed-Size"]))
            self.assertTrue(re.search("^[0-9]+$", pkg["Size"]))
        else:
            required_keys.append("installedVersion")

        for key in required_keys:
            self.assertIn(key, pkg)


    @classmethod
    def get_opkg_info(self):
        device = util.Env.device
        res = self.ssh.send_cmd(f"cat /etc/version")
        self.platform = re.search(r"^([^_]+)_", res).group(1)
        resp = requests.get(f"https://wiki.teltonika-networks.com/view/{device}_Package_Downloads")
        self.wiki_content = resp.text
        self.release_fw_version = None
        self.pkg_wiki_url = None
        self.opkg_hash = None
        m = re.search(rf"{self.platform}_R_[.\d]+", resp.text)
        if m:
            self.release_fw_version = m.group()
        m = re.search(r"https://opkg.teltonika-networks.com/([a-z0-9A-Z]+)/wiki/", resp.text)
        if m:
            self.pkg_wiki_url = m.group()
            self.opkg_hash = m.group(1)

        assert self.release_fw_version and self.pkg_wiki_url and self.opkg_hash, "Error while parsing teltonika wiki page."

        self.release_opkg_feeds_url = "http://opkg.teltonika-networks.com/" + self.opkg_hash
        resp = requests.get(self.pkg_wiki_url + TEST_PKG_TAR)
        f = open(TEST_PKG_PATH, "wb")
        f.write(resp.content)
        f.close()

    @classmethod
    def get_invalid_test_pkg(self):
        """Downloads package for other platform device (for testing returned errors)"""
        device = "RUT955" if "RUTX" in util.Env.device else "RUTX11"
        resp = requests.get(f"https://wiki.teltonika-networks.com/view/{device}_Package_Downloads")
        m = re.search(r"https://opkg.teltonika-networks.com/([a-z0-9A-Z]+)/wiki/", resp.text)
        if m:
            pkg_wiki_url = m.group()

        assert pkg_wiki_url, "Error while parsing teltonika wiki page."

        resp = requests.get(pkg_wiki_url + TEST_PKG_TAR)
        f = open(TEST_INVALID_PKG_PATH, "wb")
        f.write(resp.content)
        f.close()

    @classmethod
    def fake_versions(self):
        self.orig_fw_version = self.ssh.send_cmd("cat /etc/version").strip()
        self.orig_opkg_feeds = self.ssh.send_cmd("cat /etc/opkg/teltonikafeeds.conf").strip()

        self.ssh.send_cmd(f"echo {self.release_fw_version} > /etc/version")
        self.ssh.send_cmd(f"echo src/gz tlt_packages {self.release_opkg_feeds_url} > /etc/opkg/teltonikafeeds.conf")

    @classmethod
    def restore_versions(self):
        self.ssh.send_cmd(f"echo {self.orig_fw_version} > /etc/version")
        self.ssh.send_cmd(f"echo {self.orig_opkg_feeds} > /etc/opkg/teltonikafeeds.conf")

    # TODO Commented test for now as there are no tests for /system/device endpoint 
    # as the package endpoint has been moved to the status.lua file, when integration tests are added fot the endpoint 
    # this test should be moved also
    # def test_package_control_file_list(self):
    #     x = self.get(BASE_STATUS_URL)
    #     x.assert_code(200)
    #     data = x.json["data"]
    #     files = self.ssh.send_cmd("ls -1 /usr/lib/opkg/info/*.control | cat")
    #     for file in files.strip().split():
    #         self.assertIn(file, data)

    def test_package_list_structure(self):
        if not self.internet_connection:
            self.skipTest("Internet connection is required for this test.")

        for url in ["", "/language", "/installed", "/pending"]:
            x = self.get(BASE_CFG_URL + url)
            x.assert_code(200)

            for pkg in x.json["data"]:
                self.pkg_structure_test(pkg, url)

    def test_package_is_hidden_correctly(self):
        if not self.internet_connection:
            self.skipTest("Internet connection is required for this test.")

        hidden_pkg = []
        pkg_list = self.ssh.send_cmd("opkg-call").strip()

        for package in pkg_list.split("\r\n\r\n"):
            if "Hidden: 1" in package:
                for package_data in package.splitlines():
                    _ , success, pkg_name = package_data.partition("Package: ")
                    if success:
                        hidden_pkg.append(pkg_name)

        if len(hidden_pkg) == 0:
            self.skipTest("Hidden packages not found.")

        for url in ["", "/language", "/installed", "/pending"]:
            x = self.get(BASE_CFG_URL + url)
            x.assert_code(200)
            for pkg in x.json["data"]:
                for h_pkg in hidden_pkg:
                    self.assertNotEqual(pkg["Package"], h_pkg)

    def test_package_SHA256(self):
        if not self.internet_connection:
            self.skipTest("Internet connection is required for this test.")

        x = self.get(BASE_CFG_URL)
        x.assert_code(200)

        os.system(f"mkdir -p {TEST_PKG_DIR}")
        os.system(f"rm {TEST_PKG_DIR}/* &> /dev/null")
        os.popen(f"tar -xzvf {TEST_PKG_PATH} -C {TEST_PKG_DIR}").read()
        pkg_sha256 = os.popen(f"sha256sum {TEST_PKG_DIR}/{TEST_PKG_IPK_NAME}*.ipk").read().split()[0]

        pkg = next((pkg for pkg in x.json["data"] if pkg["Package"] == TEST_PKG_IPK_NAME), None)
        self.assertTrue(pkg)
        self.assertEqual(pkg["SHA256sum"], pkg_sha256)

    def test_package_install_offline(self):
        with self.subTest("install pkg"):
            x = self.send_file(BASE_CFG_URL, TEST_PKG_PATH)
            x.assert_code(200)
            pkg_data = x.json["data"]

            # test 'sha'
            pkg_sha256 = os.popen(f"sha256sum {TEST_PKG_PATH}").read().split()[0]
            self.assertEqual(pkg_sha256, pkg_data["sha"])

            # test 'md5'
            pkg_md5 = os.popen(f"md5sum {TEST_PKG_PATH}").read().split()[0]
            self.assertEqual(pkg_md5, pkg_data["checksum"])

            self.assertEqual(pkg_data["package"], TEST_PKG_IPK_NAME)
            self.assertEqual(pkg_data["code"], 0)
            self.assertEqual(pkg_data["verified"], True)
            self.assertEqual(pkg_data["reboot"], PKG_REBOOT)
            self.assertEqual(pkg_data["network_restart"], PKG_NETWORK_RESTART)

            x = http.post(Env.get_api_url() + INSTALL_PKG_URL, json={"data": {"custom": "1", "package": TEST_PKG_IPK_NAME}})
            self.assertEqual(x.status_code, 200)

        with self.subTest("test pkg already installed error"):
            x = self.send_file(BASE_CFG_URL, TEST_PKG_PATH)
            x.assert_code(200)

            x = http.post(Env.get_api_url() + INSTALL_PKG_URL, json={"data": {"custom": "1", "package": TEST_PKG_IPK_NAME}})
            self.assertEqual(x.status_code, 422)
            self.assertDictEqual(x.json()["errors"][0], {
                "code": 2,
                "error": "Package is already installed with same or newer version.",
                "source": "Package",
                "section": "install_package"
            })

        if self.internet_connection:
            with self.subTest("test pkg newest update error"):
                x = self.post_data(UPDATE_PKG_URL, {"package": TEST_PKG_IPK_NAME})
                x.assert_code(422)
                x.assert_error("package", "Newest package version is already installed.", 11, TEST_PKG_IPK_NAME)

            with self.subTest("test invalid pkg name update error"):
                x = self.post_data(UPDATE_PKG_URL, {"package": "invalid_pkg"})
                x.assert_code(422)
                x.assert_error("package", "Wrong package name provided.", 8, "invalid_pkg")

        with self.subTest("delete pkg"):
            x = http.post(Env.get_api_url() + DELETE_PKG_FILES_URL)
            self.assertEqual(x.status_code, 200)
            x = http.delete(Env.get_api_url() + f"{BASE_CFG_URL}/{TEST_PKG_IPK_NAME}")
            self.assertEqual(x.status_code, 200)

    def test_package_install_errors(self):
        with self.subTest("invalid file error"):
            x = self.send_file(BASE_CFG_URL, io.StringIO("invalid file content\n"))
            x.assert_code(422)
            x.assert_error("Request", "Invalid file.", 1)

        with self.subTest("missing pkg error"):
            x = self.post_data(INSTALL_PKG_URL, {"custom": "1", "package": TEST_PKG_IPK_NAME})
            x.assert_error("Package", "Missing uploaded package data.", 8)
            x.assert_code(422)

        with self.subTest("upload invalid pkg"):
            x = self.send_file(BASE_CFG_URL, TEST_INVALID_PKG_PATH)
            x.assert_code(200)

        with self.subTest("check invalid device error"):
            x = self.post_data(INSTALL_PKG_URL, {"custom": "1", "package": TEST_PKG_IPK_NAME})
            x.assert_code(422)
            x.assert_error("Package", "Not compatible with Device.", 3)

        with self.subTest("upload valid pkg"):
            x = self.send_file(BASE_CFG_URL, TEST_PKG_PATH)
            x.assert_code(200)

        with self.subTest("fake fw version"):
            self.ssh.send_cmd(f"echo invalid_version > /etc/version")
        with self.subTest("invalid fw version error"):
            x = self.post_data(INSTALL_PKG_URL, {"custom": "1", "package": TEST_PKG_IPK_NAME})
            x.assert_code(422)
            x.assert_error("Package", "Not compatible with Firmware.", 4)
        with self.subTest("reset fw version"):
            self.ssh.send_cmd(f"echo {self.release_fw_version} > /etc/version")

        with self.subTest("delete uploaded pkg"):
            x = http.post(Env.get_api_url() + DELETE_PKG_FILES_URL)
            self.assertEqual(x.status_code, 200)
            x = http.delete(Env.get_api_url() + f"{BASE_CFG_URL}/{TEST_PKG_IPK_NAME}")

    def test_package_delete_errors(self):
        x = self.delete(BASE_CFG_URL)
        x.assert_code(422)
        x.assert_error("Request", "Package name is not provided.", 122)

        x = self.delete(f"{BASE_CFG_URL}/not_found_pkg")
        x.assert_code(422)
        x.assert_error("Request", "Package not found.", 122)
    

    def test_package_install_online(self):
        if not self.internet_connection:
            self.skipTest("Internet connection is required for this test.")

        x = self.get(BASE_CFG_URL)
        x.assert_code(200)
        with self.subTest("install pkg"):
            x = http.post(Env.get_api_url() + INSTALL_PKG_URL, json={"data": {"package": TEST_PKG_IPK_NAME}})
            self.assertEqual(x.status_code, 200)
            res = self.ssh.send_cmd(f"ls -1 /usr/lib/opkg/info/{TEST_PKG_IPK_NAME}.control &> /dev/null ; echo $?")
            self.assertEqual(res.strip(), "0")

        with self.subTest("delete pkg"):
            x = http.post(Env.get_api_url() + DELETE_PKG_FILES_URL)
            self.assertEqual(x.status_code, 200)
            x = http.delete(Env.get_api_url() + f"{BASE_CFG_URL}/{TEST_PKG_IPK_NAME}")
            res = self.ssh.send_cmd(f"ls -1 /usr/lib/opkg/info/{TEST_PKG_IPK_NAME}.control &> /dev/null ; echo $?")
            self.assertNotEqual(res.strip(), "0")


    def test_package_update_online(self):
        if not self.internet_connection:
            self.skipTest("Internet connection is required for this test.")

        # using 7.1 fw version to install old pkg and test it's update to newest version
        opkg_7_1_hashes = [
                {"RUT300": "c90b38e2519a1a8a48ceb7aa46ff0dbe524773b43ce4f354fc7b1404cf0de00d"},
                {"RUT360": "e792b42385d982d12d32a79ec15671eab5b02427b83f31eff6178d5378bb1245"},
                {"RUT241": "bb408dfbf5ca9f45033f10963d3a84e3f7ee4df5686c3f9f1f38b8ced12557e8"},
                {"RUT2": "c6b0c39955e435c011556362d52b1cb2be97a57d2ccb2a14788948ab006487c1"},
                {"RUT9": "f6db7b61c3fed7591c0e464ae5fb0ebb4968d2ae876e1672480233d664a89869"},
                {"RUTX": "04d624447e31410627fa2667ba38c6f9a1c22beda543afb2ef0efbf9b5b33219"},
                {"TCR1": "4bf47903bb1b5027f43d8e250722f4356b4e84bf66ddff2dc0a99bdb49309614"},
                {"TRB1": "e83079496e9e9c35da56fad8fe2aef0533607519f2f44b673cc2a4583913e463"},
                {"TRB2": "301f0b3ac1edd3c8e893ed9f0e498911210ca0251e0ff1b83cbc7fb14dd8342f"}]

        opkg_hash = None
        for a in opkg_7_1_hashes:
            device = next(iter(a))
            hash = a[device]
            if device in util.Env.device:
                opkg_hash = hash
                break
        
        if not opkg_hash:
            self.skipTest("opkg hash not found for this test.")

        with self.subTest("fake older fw version"):
            fw_version = f"{self.platform}_R_00.07.01"
            self.ssh.send_cmd(f"echo {fw_version} > /etc/version")

            older_opkg_feeds_url = "http://opkg.teltonika-networks.com/" + opkg_hash
            self.ssh.send_cmd(f"echo src/gz tlt_packages {older_opkg_feeds_url} > /etc/opkg/teltonikafeeds.conf")
            self.fetch_server_packages()

        with self.subTest("install pkg"):
            x = http.post(Env.get_api_url() + INSTALL_PKG_URL, json={"data": {"package": TEST_PKG_IPK_NAME}})
            self.assertEqual(x.status_code, 200)


        with self.subTest("reset fw version to release"):
            self.ssh.send_cmd(f"echo {self.release_fw_version} > /etc/version")
            self.ssh.send_cmd(f"echo src/gz tlt_packages {self.release_opkg_feeds_url} > /etc/opkg/teltonikafeeds.conf")
            self.fetch_server_packages()

        with self.subTest("update pkg"):
            x = http.post(Env.get_api_url() + UPDATE_PKG_URL, json={"data": {"package": TEST_PKG_IPK_NAME}})
            self.assertEqual(x.status_code, 200)
            

        with self.subTest("delete pkg"):
            x = http.post(Env.get_api_url() + DELETE_PKG_FILES_URL)
            self.assertEqual(x.status_code, 200)
            x = http.delete(Env.get_api_url() + f"{BASE_CFG_URL}/{TEST_PKG_IPK_NAME}")

